###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This proxy provides a proxy for the MyTorchService gRPC service.
It allows the client to call specified torch operations on the server via gRPC.
"""

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch import mytorch_pb2_grpc, mytorch_pb2
from gRPC_impl import shared_msg_types_pb2
from torch.Tensor import Tensor
from torch.device import device
from utils.logger import Logger
from utils import data_transform_utils
import json
import os

from proxies.base_proxy import BaseProxy
CHUNK_SIZE = 1 * 1024 * 1024  # 1MB default

class MyTorchProxy (BaseProxy):

    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = mytorch_pb2_grpc.MyTorchServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def rand(self, *size: int):
        uuid, shape, dtype = self.generic_call("torch", "rand", *size)
        return Tensor(uuid, shape, dtype)
    
    @wrap_with_error_handler
    def randn(self, *sizes) -> Tensor:
    # Convert sizes to tuple if it's a list
        if len(sizes) == 1 and isinstance(sizes[0], list):
            sizes_tuple = tuple(sizes[0])
        else:
            sizes_tuple = sizes
            
        uuid, shape, dtype = self.generic_call("torch", "randn", sizes_tuple)
        return Tensor(uuid, shape, dtype)

    @wrap_with_error_handler
    def zeros(self, *size: int):
        uuid, shape, dtype = self.generic_call("torch", "zeros", *size)
        return Tensor(uuid, shape, dtype)
    
    @wrap_with_error_handler
    def ones(self, *size: int):
        uuid, shape, dtype = self.generic_call("torch", "ones", *size)
        return Tensor(uuid, shape, dtype)
    
    @wrap_with_error_handler
    def from_numpy(self, np_array) -> Tensor:
        request = data_transform_utils.serialize_numpy_array(np_array)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.from_numpy(request)
        tensor = Tensor(response.uuid, response.shape, response.dtype)
        tensor._retrieved_data = np_array # since we already have the data, we don't need to fetch it again
        return tensor

    @wrap_with_error_handler
    def max(self, tensor: Tensor, dim: int) -> tuple[Tensor, Tensor]:
        request = shared_msg_types_pb2.TensorIDAndDim(tensor_uuid=tensor.uuid, dim=dim)
        response: shared_msg_types_pb2.TwoGrpcTensors = self.stub.max(request)
        max_value = Tensor(response.tensor1.uuid, response.tensor1.shape, response.tensor1.dtype)
        max_indices = Tensor(response.tensor2.uuid, response.tensor2.shape, response.tensor2.dtype)
        return max_value, max_indices

    @wrap_with_error_handler
    def arange(self, start, end, step) -> Tensor:
        request = mytorch_pb2.ARangeRequest(start=start, end=end, step=step)
        response_tensor: shared_msg_types_pb2.GrpcTensor = self.stub.arange(request)
        return Tensor(response_tensor.uuid, response_tensor.shape, response_tensor.dtype)

    @wrap_with_error_handler
    def meshgrid(self, tensor_uuids, indexing) -> tuple:
        request = shared_msg_types_pb2.TensorIDsAndDim()
        request.tensor_uuids.extend(tensor_uuids)
        request.dim = indexing
        response: shared_msg_types_pb2.MultipleGrpcTensors = self.stub.meshgrid(request)
        tensors = [Tensor(tensor.uuid, tensor.shape, tensor.dtype) for tensor in response.tensors]
        return tuple(tensors)

    @wrap_with_error_handler
    def reshape(self, tensor: Tensor, shape: tuple) -> Tensor:
        request = shared_msg_types_pb2.ReshapeRequest(tensor_uuid=tensor.uuid, shape=shape)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.reshape(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    @wrap_with_error_handler
    def cat(self, tensors: list, dim: int) -> Tensor:
        request = shared_msg_types_pb2.TensorIDsAndDim()
        request.tensor_uuids.extend([tensor.uuid for tensor in tensors])
        request.dim = dim
        response: shared_msg_types_pb2.GrpcTensor = self.stub.cat(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    @wrap_with_error_handler
    def argmax(self, tensor: Tensor, dim: int, keepdim: bool) -> Tensor:
        request = mytorch_pb2.ArgMaxRequest(tensor_uuid=tensor.uuid, dim=dim, keepdim=keepdim)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.argmax(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    @wrap_with_error_handler
    def matmul(self, tensor1: Tensor, tensor2: Tensor) -> Tensor:
        uuid, shape, dtype = self.generic_call("torch", "matmul", tensor1.uuid, tensor2.uuid)
        return Tensor(uuid, shape, dtype)
    
    @wrap_with_error_handler
    def device(self, dev: str) -> device:
        returned_dev = self.generic_call("torch", "device", dev)
        return device(returned_dev)


    @wrap_with_error_handler
    def save(self, state_dict: dict, file_path: str):
        named_tensor_uuids = {
            k: v.uuid for k, v in state_dict.items() if isinstance(v, Tensor)
        }
        request = mytorch_pb2.SaveModelRequest(
            filename=os.path.basename(file_path),
            named_tensor_uuids=named_tensor_uuids
        )
        chunk_stream = self.stub.save_model_stream(request)

        with open(file_path, "wb") as f:
            for chunk in chunk_stream:
                f.write(chunk.chunk_data)

    @wrap_with_error_handler
    def load(self, file_path: str):
        filename = os.path.basename(file_path)
        def chunk_generator():
            with open(file_path, "rb") as f:
                index = 0
                while True:
                    chunk = f.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    yield shared_msg_types_pb2.FileChunk(
                        filename=filename,
                        chunk_data=chunk,
                        chunk_index=index,
                        is_last_chunk=False
                    )
                    index += 1
                yield shared_msg_types_pb2.FileChunk(
                    filename=filename,
                    chunk_data=b"",
                    chunk_index=index,
                    is_last_chunk=True
                )

        response = self.stub.load_model_stream(chunk_generator())
        return {
            name: Tensor(t.uuid, list(t.shape), t.dtype)
            for name, t in response.tensors.items()
        }

    @wrap_with_error_handler
    def allclose(self, input: Tensor, other: Tensor, rtol: float = 1e-05, atol: float = 1e-08, equal_nan: bool = False) -> bool:
        request = mytorch_pb2.AllCloseRequest(tensor1_uuid=input.uuid, tensor2_uuid=other.uuid, rtol=rtol, atol=atol, equal_nan=equal_nan)
        response = self.stub.allclose(request)
        return response.value

    @wrap_with_error_handler
    def unsqueeze(self, tensor: Tensor, dim: int) -> Tensor:
        request = shared_msg_types_pb2.TensorIDAndDim(tensor_uuid=tensor.uuid, dim=dim)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.unsqueeze(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    def repeat_interleave(self, tensor: Tensor, repeats: int, dim: int) -> Tensor:
        request = mytorch_pb2.RepeatInterleaveRequest(tensor_uuid=tensor.uuid, repeats=repeats, dim=dim)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.repeat_interleave(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    @wrap_with_error_handler
    def tanh(self, tensor: Tensor) -> Tensor:
        uuid, shape, dtype = self.generic_call("torch", "tanh", tensor.uuid)
        return Tensor(uuid, shape, dtype)


    @wrap_with_error_handler
    def sigmoid(self, tensor: Tensor) -> Tensor:
        uuid, shape, dtype = self.generic_call("torch", "sigmoid", tensor.uuid)
        return Tensor(uuid, shape, dtype)


    @wrap_with_error_handler    
    def stack(self, tensor_uuids, dim: int = 0):
        uuid, shape, dtype  = self.generic_call("torch", "stack", tensor_uuids, dim)
        return Tensor(uuid, shape, dtype)

    @wrap_with_error_handler
    def mean(self, tensor: Tensor, dim=None, keepdim=False, in_dtype=None, out=None) -> Tensor:
        if dim is None and keepdim is False and in_dtype is None and out is None:
            uuid, shape, dtype = self.generic_call("torch", "mean", tensor.uuid)
        elif dim is not None and keepdim is False and in_dtype is None and out is None:
            uuid, shape, dtype = self.generic_call("torch", "mean", tensor.uuid, dim)
        else:
            uuid, shape, dtype = self.generic_call("torch", "mean", tensor.uuid, dim, keepdim, in_dtype, out)
        return Tensor(uuid, shape, dtype)


    @wrap_with_error_handler
    def equal(self, tensor1: Tensor, tensor2: Tensor) -> bool:
        yesno = self.generic_call("torch", "equal", tensor1.uuid, tensor2.uuid)
        return yesno
    
    ### CACHEQ methods:

    @wrap_with_error_handler
    def scaffolding_server_get_timing_statistics(self, run_id: str) -> str:
        request = shared_msg_types_pb2.StringValue()
        request.value = run_id
        return self.stub.scaffolding_server_get_timing_statistics(request) 
    
    @wrap_with_error_handler
    def scaffolding_server_initialize_timing_statistics(self, run_id: str):
        request = shared_msg_types_pb2.StringValue()
        request.value = run_id
        self.stub.scaffolding_server_initialize_timing_statistics(request)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from .Transformer import Transformer
from proxies.mytorchvision.transforms.ToTensorProxy import ToTensorProxy

class ToTensor (Transformer):
    def __init__(self):
        super().__init__()
        self.proxy = ToTensorProxy()
        self.uuid = self.proxy.createTransformOnServer()




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/optim/optimizers.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1emytorch/optim/optimizers.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"\\\n\x19\x43reateSGDOptimizerRequest\x12\x16\n\x0egenerator_uuid\x18\x01 \x01(\t\x12\x15\n\rlearning_rate\x18\x02 \x01(\x02\x12\x10\n\x08momentum\x18\x03 \x01(\x02*|\n\rOptimizerType\x12\x07\n\x03SGD\x10\x00\x12\x08\n\x04\x41\x44\x41M\x10\x01\x12\x0b\n\x07\x41\x44\x41GRAD\x10\x02\x12\x0b\n\x07RMSPROP\x10\x03\x12\n\n\x06\x41\x44\x41MAX\x10\x04\x12\t\n\x05\x41\x44\x41MW\x10\x05\x12\x0b\n\x07\x41\x44\x41MAXW\x10\x06\x12\x0c\n\x08\x41\x44\x41GRADW\x10\x07\x12\x0c\n\x08RMSPROPW\x10\x08\x32\xa8\x01\n\x10OptimizerService\x12\x46\n\x12\x43reateSGDOptimizer\x12\".mytorch.CreateSGDOptimizerRequest\x1a\x0c.shared.UUID\x12\'\n\x08ZeroGrad\x12\x0c.shared.UUID\x1a\r.shared.Empty\x12#\n\x04Step\x12\x0c.shared.UUID\x1a\r.shared.Emptyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.optim.optimizers_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_OPTIMIZERTYPE']._serialized_start=161
  _globals['_OPTIMIZERTYPE']._serialized_end=285
  _globals['_CREATESGDOPTIMIZERREQUEST']._serialized_start=67
  _globals['_CREATESGDOPTIMIZERREQUEST']._serialized_end=159
  _globals['_OPTIMIZERSERVICE']._serialized_start=288
  _globals['_OPTIMIZERSERVICE']._serialized_end=456
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytransformers/HuggingFaceTokenizer.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)mytransformers/HuggingFaceTokenizer.proto\x12\x0emytransformers\x1a\x16shared_msg_types.proto\"`\n\x12\x65ncodePlus_request\x12\x16\n\x0etokenizer_uuid\x18\x01 \x01(\t\x12\x0c\n\x04text\x18\x02 \x01(\t\x12$\n\x06kwargs\x18\x03 \x03(\x0b\x32\x14.shared.keyValuePair\";\n\x17tokenizer_key_uuid_pair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x13\n\x0btensor_uuid\x18\x02 \x01(\t\"c\n\x13\x65ncodePlus_response\x12\x15\n\rencoding_uuid\x18\x01 \x01(\t\x12\x35\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\'.mytransformers.tokenizer_key_uuid_pair\"@\n\x13setPadToken_request\x12\x16\n\x0etokenizer_uuid\x18\x01 \x01(\t\x12\x11\n\tpad_token\x18\x02 \x01(\t\"\x16\n\x14setPadToken_response\"i\n\x0e\x64\x65\x63ode_request\x12\x16\n\x0etokenizer_uuid\x18\x01 \x01(\t\x12\x19\n\x11input_tensor_uuid\x18\x02 \x01(\t\x12$\n\x06kwargs\x18\x03 \x03(\x0b\x32\x14.shared.keyValuePair\"\x1f\n\x0f\x64\x65\x63ode_response\x12\x0c\n\x04text\x18\x01 \x01(\t\"\xd4\x01\n\x1ftokenizerFromPretrained_request\x12%\n\x1dpretrained_model_name_or_path\x18\x01 \x01(\t\x12\x0e\n\x06inputs\x18\x02 \x03(\t\x12K\n\x06kwargs\x18\x03 \x03(\x0b\x32;.mytransformers.tokenizerFromPretrained_request.KwargsEntry\x1a-\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xcd\x01\n tokenizerFromPretrained_response\x12\x16\n\x0etokenizer_uuid\x18\x01 \x01(\t\x12[\n\x0especial_tokens\x18\x02 \x03(\x0b\x32\x43.mytransformers.tokenizerFromPretrained_response.SpecialTokensEntry\x1a\x34\n\x12SpecialTokensEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\xf7\x06\n\x1bHuggingFaceTokenizerService\x12u\n\x10getAutoTokenizer\x12/.mytransformers.tokenizerFromPretrained_request\x1a\x30.mytransformers.tokenizerFromPretrained_response\x12u\n\x10getBertTokenizer\x12/.mytransformers.tokenizerFromPretrained_request\x1a\x30.mytransformers.tokenizerFromPretrained_response\x12x\n\x13getRobertaTokenizer\x12/.mytransformers.tokenizerFromPretrained_request\x1a\x30.mytransformers.tokenizerFromPretrained_response\x12s\n\x0egetT5Tokenizer\x12/.mytransformers.tokenizerFromPretrained_request\x1a\x30.mytransformers.tokenizerFromPretrained_response\x12v\n\x11getXLNetTokenizer\x12/.mytransformers.tokenizerFromPretrained_request\x1a\x30.mytransformers.tokenizerFromPretrained_response\x12X\n\x0b\x65ncode_plus\x12\".mytransformers.encodePlus_request\x1a#.mytransformers.encodePlus_response\"\x00\x12\\\n\rset_pad_token\x12#.mytransformers.setPadToken_request\x1a$.mytransformers.setPadToken_response\"\x00\x12K\n\x06\x64\x65\x63ode\x12\x1e.mytransformers.decode_request\x1a\x1f.mytransformers.decode_response\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytransformers.HuggingFaceTokenizer_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TOKENIZERFROMPRETRAINED_REQUEST_KWARGSENTRY']._loaded_options = None
  _globals['_TOKENIZERFROMPRETRAINED_REQUEST_KWARGSENTRY']._serialized_options = b'8\001'
  _globals['_TOKENIZERFROMPRETRAINED_RESPONSE_SPECIALTOKENSENTRY']._loaded_options = None
  _globals['_TOKENIZERFROMPRETRAINED_RESPONSE_SPECIALTOKENSENTRY']._serialized_options = b'8\001'
  _globals['_ENCODEPLUS_REQUEST']._serialized_start=85
  _globals['_ENCODEPLUS_REQUEST']._serialized_end=181
  _globals['_TOKENIZER_KEY_UUID_PAIR']._serialized_start=183
  _globals['_TOKENIZER_KEY_UUID_PAIR']._serialized_end=242
  _globals['_ENCODEPLUS_RESPONSE']._serialized_start=244
  _globals['_ENCODEPLUS_RESPONSE']._serialized_end=343
  _globals['_SETPADTOKEN_REQUEST']._serialized_start=345
  _globals['_SETPADTOKEN_REQUEST']._serialized_end=409
  _globals['_SETPADTOKEN_RESPONSE']._serialized_start=411
  _globals['_SETPADTOKEN_RESPONSE']._serialized_end=433
  _globals['_DECODE_REQUEST']._serialized_start=435
  _globals['_DECODE_REQUEST']._serialized_end=540
  _globals['_DECODE_RESPONSE']._serialized_start=542
  _globals['_DECODE_RESPONSE']._serialized_end=573
  _globals['_TOKENIZERFROMPRETRAINED_REQUEST']._serialized_start=576
  _globals['_TOKENIZERFROMPRETRAINED_REQUEST']._serialized_end=788
  _globals['_TOKENIZERFROMPRETRAINED_REQUEST_KWARGSENTRY']._serialized_start=743
  _globals['_TOKENIZERFROMPRETRAINED_REQUEST_KWARGSENTRY']._serialized_end=788
  _globals['_TOKENIZERFROMPRETRAINED_RESPONSE']._serialized_start=791
  _globals['_TOKENIZERFROMPRETRAINED_RESPONSE']._serialized_end=996
  _globals['_TOKENIZERFROMPRETRAINED_RESPONSE_SPECIALTOKENSENTRY']._serialized_start=944
  _globals['_TOKENIZERFROMPRETRAINED_RESPONSE_SPECIALTOKENSENTRY']._serialized_end=996
  _globals['_HUGGINGFACETOKENIZERSERVICE']._serialized_start=999
  _globals['_HUGGINGFACETOKENIZERSERVICE']._serialized_end=1886
# @@protoc_insertion_point(module_scope)



from gRPC_impl.mytorch.scaffolding import data_mgmt_pb2, data_mgmt_pb2_grpc
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from utils.logger import Logger
import os
from pathlib import Path
from tqdm import tqdm
import time

class DataMgmtProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = data_mgmt_pb2_grpc.DataMgmtServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def folder_exists_on_server(self, relative_path) -> bool:
        response = self.stub.FolderExistsOnServer(data_mgmt_pb2.RelativePath(path=relative_path))
        return response.value

    @wrap_with_error_handler
    def upload_folder(self, folder_path) -> int:
        """
        Uploads the specified folder and its contents, including subfolders.

        For example, if the folder path is '/home/Users/gduda/mytorch/test_data/resnet_infer', the folder 'resnet_infer' and all its contents
        will be uploaded to the server. On the server, the folder `resnet_infer` will be created in the base path where files are saved.

        Args:
            folder_path (str): The path to the folder to upload, can be relative (from the current working directory).
        """
        # Resolve absolute path from the relative path (if given) and ensure the base path is the parent directory
        folder_path = os.path.abspath(folder_path)
        base_path = os.path.dirname(folder_path)  # Use the parent directory as the base
        trailing_folder = os.path.basename(folder_path)
        self.logger.info(f"Uploading folder: {folder_path} to <server data dir>/{trailing_folder} on server")

        # Collect all files to upload
        files_to_upload = [(file_path.as_posix(), os.path.relpath(file_path.as_posix(), start=base_path))
                           for root, _, files in os.walk(folder_path)
                           for file_path in [Path(root).joinpath(filename) for filename in files]]

        start_time = time.time()  # Start timing the upload process

        # Initialize tqdm progress bar
        with tqdm(total=len(files_to_upload), unit='files', desc=f"Uploading Files from {folder_path}",
                  bar_format="{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt}") as progress_bar:
            def generate_file_stream():
                for file_path, relative_path in files_to_upload:
                    with open(file_path, 'rb') as file:
                        content = file.read()
                        yield data_mgmt_pb2.FileUploadRequest(filepath=relative_path, content=content)
                    progress_bar.update(1)  # Update progress bar per file uploaded

            # Perform the upload and get response
            response = self.stub.UploadFolder(generate_file_stream())

        # Calculate total time taken
        total_time = time.time() - start_time

        # Print final message with total files uploaded and total time taken
        self.logger.info(f"Total files uploaded: {response.count}")
        self.logger.info(f"Total time taken: {total_time:.2f} seconds")
        return response.count


from gRPC_impl.mytorch.scaffolding import resource_mgmt_pb2_grpc, resource_mgmt_pb2
from connection_utils.server_connection import ServerConnection
from utils.logger import Logger

class ResourceMgmtProxy:
    def __init__(self, grpc_channel):
        self.stub = resource_mgmt_pb2_grpc.ResourceMgmtServiceStub(grpc_channel)
        self.logger = Logger.get_logger()

    def resource_request(self, auth_token: str, client_ip: str, server_id: str, max_time: int):
        request = resource_mgmt_pb2.ResourceRequest()
        request.auth_token = auth_token
        request.client_ip = client_ip
        request.server_id = server_id
        request.max_time = max_time
        response: resource_mgmt_pb2.ResourceRequestResponse = self.stub.resource_request(request)
        self.logger.info(f"Received response from Resource Manager. Resource ID: {response.resource_id if response.resource_id else 'None available'}.")
        # return response as a dictionary
        response_dict = {
            "access_token": response.access_token,
            "additional_info": response.additional_info,
            "resource_id": response.resource_id,
            "resource_ip": response.resource_ip,
            "resource_port": response.resource_port,
            "resource_name": response.resource_name,
            "lease_time": response.lease_time,
            "mytorch_version": response.mytorch_version,
            "resource_info": response.resource_info
        }
        return response_dict



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

class BatchEncoding:
    def __init__(self, uuid, data_dict):
        self.uuid = uuid
        self.data = data_dict # {string: Tensor}

    # have the ability to access the data dictionary as if it were an attribute
    def __getitem__(self, key):
        return self.data[key]

    def items(self):
        return self.data.items()

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def get(self, key, default=None):
        return self.data.get(key, default)

    def __contains__(self, key):
        return key in self.data

    def to_dict(self):
        return dict(self.data)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/tensor.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14mytorch/tensor.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"\x93\x02\n\x16TensorStreamingRequest\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x1c\n\x04uuid\x18\x02 \x01(\x0b\x32\x0c.shared.UUIDH\x00\x12+\n\x0btwo_tensors\x18\x03 \x01(\x0b\x32\x14.shared.TwoTensorIDsH\x00\x12\x30\n\x0etensor_and_dim\x18\x04 \x01(\x0b\x32\x16.shared.TensorIDAndDimH\x00\x12\x31\n\x0freshape_request\x18\x05 \x01(\x0b\x32\x16.shared.ReshapeRequestH\x00\x12.\n\rslice_request\x18\x06 \x01(\x0b\x32\x15.mytorch.SliceRequestH\x00\x42\t\n\x07payload\"\xc9\x01\n\x17TensorStreamingResponse\x12$\n\x06tensor\x18\x01 \x01(\x0b\x32\x12.shared.GrpcTensorH\x00\x12)\n\x0b\x66loat_value\x18\x02 \x01(\x0b\x32\x12.shared.FloatValueH\x00\x12+\n\x0ctensor_shape\x18\x03 \x01(\x0b\x32\x13.shared.TensorShapeH\x00\x12%\n\x0ctensor_dtype\x18\x04 \x01(\x0b\x32\r.shared.DtypeH\x00\x42\t\n\x07payload\"8\n\x0bTensorSlice\x12\r\n\x05start\x18\x01 \x01(\x05\x12\x0c\n\x04stop\x18\x02 \x01(\x05\x12\x0c\n\x04step\x18\x03 \x01(\x05\"I\n\x0cSliceRequest\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t\x12$\n\x06slices\x18\x02 \x03(\x0b\x32\x14.mytorch.TensorSlice2\xcb\x07\n\rTensorService\x12)\n\x05\x66loat\x12\x0c.shared.UUID\x1a\x12.shared.GrpcTensor\x12(\n\x04long\x12\x0c.shared.UUID\x1a\x12.shared.GrpcTensor\x12,\n\x08\x62\x61\x63kward\x12\x0c.shared.UUID\x1a\x12.shared.GrpcTensor\x12+\n\x07to_cuda\x12\x0c.shared.UUID\x1a\x12.shared.GrpcTensor\x12*\n\x06to_cpu\x12\x0c.shared.UUID\x1a\x12.shared.GrpcTensor\x12\'\n\x03sum\x12\x0c.shared.UUID\x1a\x12.shared.GrpcTensor\x12/\n\x03\x61\x64\x64\x12\x14.shared.TwoTensorIDs\x1a\x12.shared.GrpcTensor\x12\x31\n\x05\x65qual\x12\x14.shared.TwoTensorIDs\x1a\x12.shared.GrpcTensor\x12\x33\n\x05index\x12\x16.shared.TensorIDAndDim\x1a\x12.shared.GrpcTensor\x12\x37\n\tunsqueeze\x12\x16.shared.TensorIDAndDim\x1a\x12.shared.GrpcTensor\x12\x35\n\x07reshape\x12\x16.shared.ReshapeRequest\x1a\x12.shared.GrpcTensor\x12\x32\n\x05slice\x12\x15.mytorch.SliceRequest\x1a\x12.shared.GrpcTensor\x12\x36\n\x08get_data\x12\x0c.shared.UUID\x1a\x1c.shared.SerializedTensorData\x12(\n\x04item\x12\x0c.shared.UUID\x1a\x12.shared.FloatValue\x12.\n\tget_shape\x12\x0c.shared.UUID\x1a\x13.shared.TensorShape\x12(\n\tget_dtype\x12\x0c.shared.UUID\x1a\r.shared.Dtype\x12%\n\x06\x64\x65lete\x12\x0c.shared.UUID\x1a\r.shared.Empty\x12Z\n\x11TensorStreamingOp\x12\x1f.mytorch.TensorStreamingRequest\x1a .mytorch.TensorStreamingResponse(\x01\x30\x01\x12\x39\n\x0cgeneric_call\x12\x13.shared.JsonRequest\x1a\x14.shared.JsonResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.tensor_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TENSORSTREAMINGREQUEST']._serialized_start=58
  _globals['_TENSORSTREAMINGREQUEST']._serialized_end=333
  _globals['_TENSORSTREAMINGRESPONSE']._serialized_start=336
  _globals['_TENSORSTREAMINGRESPONSE']._serialized_end=537
  _globals['_TENSORSLICE']._serialized_start=539
  _globals['_TENSORSLICE']._serialized_end=595
  _globals['_SLICEREQUEST']._serialized_start=597
  _globals['_SLICEREQUEST']._serialized_end=670
  _globals['_TENSORSERVICE']._serialized_start=673
  _globals['_TENSORSERVICE']._serialized_end=1644
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.optim import optimizers_pb2_grpc, optimizers_pb2
from gRPC_impl import shared_msg_types_pb2
from utils.logger import Logger

from proxies.base_proxy import BaseProxy

class OptimizerProxy(BaseProxy):
    def __init__(self):
        super().__init__()

        self.channel = ServerConnection.get_active_connection()
        self.stub = optimizers_pb2_grpc.OptimizerServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = None

    @wrap_with_error_handler
    def create_SGD(self, generator_uuid: str, learning_rate: float, momentum):
        request = optimizers_pb2.CreateSGDOptimizerRequest()
        request.generator_uuid = generator_uuid
        request.learning_rate = learning_rate
        request.momentum = momentum
        response = self.stub.CreateSGDOptimizer(request)
        return response.uuid

    @wrap_with_error_handler
    def zero_grad(self):
        uuid_msg = shared_msg_types_pb2.UUID(uuid=self.uuid)
        return self.stub.ZeroGrad(uuid_msg)

    @wrap_with_error_handler
    def step(self):
        uuid_msg = shared_msg_types_pb2.UUID(uuid=self.uuid)
        return self.stub.Step(uuid_msg)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torch.Tensor import Tensor
from typing import List, Union
from numpy import ndarray
from PIL import Image
from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorchvision.transforms import torchvision_functional_pb2, torchvision_functional_pb2_grpc
from gRPC_impl import shared_msg_types_pb2
import numpy as np
from utils import data_transform_utils

class TorchvisionFunctionalProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = torchvision_functional_pb2_grpc.TorchvisionFunctionalServiceStub(self.channel)
        self.logger = Logger.get_logger()

    # same as torchvision.transforms.functional.resize
    @wrap_with_error_handler
    def resize(self, img: Tensor, size: List[int]) -> Tensor:
        request = torchvision_functional_pb2.ResizeRequest()
        request.tensor_uuid = img.uuid
        request.size.extend(size)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.resize(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    # same as torchvision.transforms.functional.to_tensor
    @wrap_with_error_handler
    def to_tensor(self, pic: Union[Image, ndarray]) -> Tensor:
        if isinstance(pic, Image.Image):
            np_array = np.array(pic)
        else:
            np_array = pic
        request: shared_msg_types_pb2.SerializedNumpyArray = data_transform_utils.serialize_numpy_array(np_array)
        response: shared_msg_types_pb2.GrpcTensor = self.stub.to_tensor(request)
        return Tensor(response.uuid, response.shape, response.dtype)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.mytorch.optim.optimizer_proxy import OptimizerProxy
from torch.nn.ParametersGenerator import ParametersGenerator

class SGD:
    def __init__(self, generator: ParametersGenerator, lr: float, momentum: float = 0):
        self.proxy = OptimizerProxy()
        self.uuid = self.proxy.create_SGD(generator.uuid, lr, momentum)
        self.proxy.uuid = self.uuid

    def zero_grad(self):
        self.proxy.zero_grad()

    def step(self):
        self.proxy.step()


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from torch.utils.data.Dataset import Dataset
from proxies.base_proxy import BaseProxy
from torch.Tensor import Tensor  
from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler

class DataLoaderProxy(BaseProxy):
    def __init__(self):
        super().__init__()
        self.logger = Logger.get_logger()
        self.channel = ServerConnection.get_active_connection()

    def createDataLoaderOnServer(self, dataset: Dataset, batch_size: int, shuffle: bool):
        self.uuid, dataset_length, batch_size = self.generic_call("torch.utils.data", 
                                                                  "DataLoader", 
                                                                  dataset.uuid, 
                                                                  batch_size, 
                                                                  shuffle,
                                                                  call_type="constructor")
        return self.uuid, dataset_length, batch_size

    def _deserialize_tensor(self, tup):
        """Convert a (uuid, shape, dtype) tuple to a MyTorch Tensor."""
        if isinstance(tup, (list, tuple)) and len(tup) == 3:
            uuid, shape, dtype = tup
            return Tensor(uuid, shape, dtype)
        return tup  # Return raw if it's not a tensor-like tuple

    def __iter__(self):
        """Implements: for batch in dataloader"""
        batches = self.generic_call("torch.utils.data", "__iter__", self.uuid, call_type="method")

        def convert(batch):
            if isinstance(batch, dict):  # Optional support
                return (
                    self._deserialize_tensor(batch.get("inputs")),
                    self._deserialize_tensor(batch.get("targets"))
                )
            if isinstance(batch, (list, tuple)):
                return tuple(self._deserialize_tensor(x) for x in batch)
            return self._deserialize_tensor(batch)

        return iter([convert(batch) for batch in batches])

    def __getitem__(self, index):
        batch = self.generic_call("torch.utils.data", "__getitem__", self.uuid, index)
        if isinstance(batch, (list, tuple)):
            return tuple(self._deserialize_tensor(x) for x in batch)
        return self._deserialize_tensor(batch)

__all__ = ['DataLoaderProxy']



###############################################################################
# Copyright (c) 2018-2019 MyTorch Systems Inc. All rights reserved.
###############################################################################
"""
This file contains a NoOpContextManager class that is used when we want
to execute a block of code with no side effects. This is useful when we
want to use a with statement to execute a block of code such as:

    with torch.no_grad():
        blah
        blah
        blah

"""

class NoOpContextManager:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

'''
Because all of the HuggingFace models have the same API, we can create a single proxy class that can be used to interact with any of them.
'''

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl import shared_msg_types_pb2
from gRPC_impl.mytransformers import HuggingFaceModel_pb2_grpc, HuggingFaceModel_pb2
from utils.logger import Logger
from utils.data_transform_utils import convert_grpc_kwargs_to_python, convert_python_to_grpc_kwargs
from torch.Tensor import Tensor

class HuggingFaceModelProxy:

    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = HuggingFaceModel_pb2_grpc.HuggingFaceModelServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def getAutoModelForSequenceClassification(self, pretrained_model_name_or_path, *model_args, **kwargs):
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs)
        response = self.stub.getAutoModelForSequenceClassification(request)
        return response.model_uuid

    @wrap_with_error_handler
    def getAutoModelForCausalML(self, pretrained_model_name_or_path, *model_args, **kwargs):
        # Convert all kwargs values to gRPC Value types
        kwargs_grpc = convert_python_to_grpc_kwargs(kwargs)
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs_grpc)
        response = self.stub.getAutoModelForCausalML(request)
        return response.model_uuid

    @wrap_with_error_handler
    def getBertModelForSequenceClassification(self, pretrained_model_name_or_path, *model_args, **kwargs):
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs)
        response = self.stub.getBertModelForSequenceClassification(request)
        return response.model_uuid

    @wrap_with_error_handler
    def getRobertaModelForSequenceClassification(self, pretrained_model_name_or_path, *model_args, **kwargs):
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs)
        response = self.stub.getRobertaModelForSequenceClassification(request)
        return response.model_uuid

    @wrap_with_error_handler
    def getT5ForConditionalGeneration(self, pretrained_model_name_or_path, *model_args, **kwargs):
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs)
        response = self.stub.getT5ForConditionalGeneration(request)
        return response.model_uuid

    @wrap_with_error_handler
    def getAutoModelForSeq2SeqLM(self, pretrained_model_name_or_path, *model_args, **kwargs):
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs)
        response = self.stub.getAutoModelForSeq2SeqLM(request)
        return response.model_uuid

    @wrap_with_error_handler
    def getXLNetForSequenceClassification(self, pretrained_model_name_or_path, *model_args, **kwargs):
        request = HuggingFaceModel_pb2.fromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         model_args=model_args,
                                                         kwargs=kwargs)
        response = self.stub.getXLNetForSequenceClassification(request)
        return response.model_uuid

    @wrap_with_error_handler
    def generate(self, model_uuid, **kwargs):
        request = HuggingFaceModel_pb2.generate_request(model_uuid=model_uuid)
        for key, value in kwargs.items():
            if isinstance(value, Tensor):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, string_value=value.uuid)])
            elif isinstance(value, bool):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, bool_value=value)])
            elif isinstance(value, int):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, int_value=value)])
            elif isinstance(value, float):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, float_value=value)])
        response = self.stub.generate(request)
        return response.tensor_uuid


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.nn.functional import functional_pb2, functional_pb2_grpc
from torch.Tensor import Tensor
from utils.logger import Logger

class FunctionalProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = functional_pb2_grpc.FunctionalServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = None

    @wrap_with_error_handler
    def softmax(self, input: Tensor, dim: int) -> Tensor:
        request = functional_pb2.SoftmaxRequest(tensor_uuid=input.uuid, dim=dim)
        response = self.stub.softmax(request)
        return Tensor(response.uuid, response.shape, response.dtype)
    
    
    #@wrap_with_error_handler
    #def relu(self, input: Tensor, dim: int) -> Tensor:
    #    return self.generic_call("torch", "rand", *size) # TODO Make just ONE generic_call, currently many


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/hub/hub.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15mytorch/hub/hub.proto\x12\x07mytorch\"J\n\x10ModelLoadRequest\x12\x0c\n\x04repo\x18\x01 \x01(\t\x12\x12\n\nmodel_name\x18\x02 \x01(\t\x12\x14\n\x0cweights_enum\x18\x03 \x01(\t\"\'\n\x11ModelLoadResponse\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t2P\n\nHubService\x12\x42\n\tloadModel\x12\x19.mytorch.ModelLoadRequest\x1a\x1a.mytorch.ModelLoadResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.hub.hub_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MODELLOADREQUEST']._serialized_start=34
  _globals['_MODELLOADREQUEST']._serialized_end=108
  _globals['_MODELLOADRESPONSE']._serialized_start=110
  _globals['_MODELLOADRESPONSE']._serialized_end=149
  _globals['_HUBSERVICE']._serialized_start=151
  _globals['_HUBSERVICE']._serialized_end=231
# @@protoc_insertion_point(module_scope)



from gRPC_impl.mytorch.scaffolding import server_mgmt_pb2, server_mgmt_pb2_grpc
from gRPC_impl import shared_msg_types_pb2
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from utils.logger import Logger

class ServerMgmtProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = server_mgmt_pb2_grpc.ServerMgmtServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def server_status(self) -> str:
        response: server_mgmt_pb2.ServerStatus = self.stub.server_status(shared_msg_types_pb2.Empty())
        return response.server_status

    # called when the client disconnects: either gracefully or due to an error or forceful termination
    def client_disconnect(self):
        self.stub.client_disconnect(shared_msg_types_pb2.Empty())

    def get_server_gpu_stats(self) -> [dict]:
        response: server_mgmt_pb2.GpuInfoList = self.stub.get_server_gpu_stats(shared_msg_types_pb2.Empty())
        gpu_stats = []
        for gpu_info_dict in response.gpu_dicts:
            gpu_stats.append(gpu_info_dict.items)
        return gpu_stats


from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class SoftmaxRequest(_message.Message):
    __slots__ = ("tensor_uuid", "dim")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    DIM_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    dim: int
    def __init__(self, tensor_uuid: _Optional[str] = ..., dim: _Optional[int] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torch.utils.data.Dataset import Dataset
from typing import Sequence
from proxies.mytorch.utils.data.subset_proxy import SubsetProxy

class Subset(Dataset):
    def __init__(self, dataset: Dataset, indices: Sequence[int]) -> None:
        super().__init__()
        self.dataset = dataset
        self.indices = indices
        self.proxy = SubsetProxy()
        self.uuid, self.dataset_length = self.createSubsetOnServer()

    def createSubsetOnServer(self):
        uuid, dataset_length = self.proxy.createSubsetOnServer(self.dataset, self.indices)
        return uuid, dataset_length

    def __len__(self):
        return self.dataset_length


###############################################################################
# File: mytorch/client/proxies/mytorch/nn/lstm_proxy.py
###############################################################################
"""
Proxy that represents a server-side LSTM “Module.”
We store an LSTM on the server via generic_call(…).
"""
from proxies.mytorch.nn.module_proxy import ModuleProxy
from torch.Tensor import Tensor

class LSTMProxy(ModuleProxy):
    """
    Re-uses all the “ModuleProxy” logic so it can store parameters, do forward, etc.
    """

    def create_lstm_on_server(self, input_size, hidden_size, num_layers, bias,
                              batch_first, dropout, bidirectional):
        """
        Tells the server to create a real torch.nn.LSTM(...) object, store it,
        and return its UUID for tracking.
        """
        # We’ll pass these as arguments to a new “context=nn_module, method=create_lstm_module.”
        # The server side will interpret these, create LSTM, store it, return {uuid=...}.
        uuid = self.generic_call(
            "torch.nn", 
            "LSTM",
            input_size,
            hidden_size,
            num_layers,
            bias,
            batch_first,
            dropout,
            bidirectional,
            call_type="constructor"
        )
        # The server will return e.g. {"type": "module", "uuid": "..."} 
        return uuid
    
    def forward(self, module_uuid: str, input_tensor: Tensor, hx=None):
        """
        This calls the existing server reflection to do module.forward(input_tensor, hx).
        The server now returns a nested list: [ (out_uuid, out_shape, out_dtype),
                                              [ (h_uuid, h_shape, h_dtype),
                                                (c_uuid, c_shape, c_dtype) ] ]
        We'll parse it accordingly.
        """
        args = [module_uuid, input_tensor.uuid]
        if hx is not None:
            # hx = (h0, c0)
            h_uuid = hx[0].uuid
            c_uuid = hx[1].uuid
            args.extend([h_uuid, c_uuid])

        # generic_call returns a dict describing output, h_n, c_n, etc.
        # Something like:
        # {
        #   "output": ["uuid_for_output", [shape], "torch.float32"],
        #   "h_n":    ["uuid_for_hn",     [shape], "torch.float32"],
        #   "c_n":    ["uuid_for_cn",     [shape], "torch.float32"]
        # }
        raw_result = self.generic_call("torch.nn", "forward", *args, call_type="forward")


        # raw_result is presumably a list of length 2:
        # [
        #   (out_uuid, out_shape, out_dtype),
        #   [
        #     (h_uuid, h_shape, h_dtype),
        #     (c_uuid, c_shape, c_dtype)
        #   ]
        # ]
        # We'll parse it carefully:
        out_info, hc_info = raw_result
        # out_info = (out_uuid, out_shape, out_dtype)
        (out_uuid, out_shape, out_dtype) = out_info

        # hc_info is a list: [ (h_uuid,h_shape,h_dtype), (c_uuid,c_shape,c_dtype) ]
        h_info, c_info = hc_info
        (h_uuid, h_shape, h_dtype) = h_info
        (c_uuid, c_shape, c_dtype) = c_info

        # Now we can build MyTorch Tensor objects
        from torch.Tensor import Tensor  # local import to avoid circular references
        out_tensor = Tensor(out_uuid, out_shape, out_dtype)
        h_n = Tensor(h_uuid, h_shape, h_dtype)
        c_n = Tensor(c_uuid, c_shape, c_dtype)

        return (out_tensor, (h_n, c_n))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/scaffolding/server_mgmt.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n mytorch/scaffolding/server_mgmt.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"b\n\x0cServerStatus\x12\x13\n\x0bserver_name\x18\x01 \x01(\t\x12\x11\n\tserver_ip\x18\x02 \x01(\t\x12\x13\n\x0bserver_port\x18\x03 \x01(\x05\x12\x15\n\rserver_status\x18\x04 \x01(\t\":\n\x0bGpuInfoList\x12+\n\tgpu_dicts\x18\x01 \x03(\x0b\x32\x18.shared.StringDictionary2\xbc\x01\n\x11ServerMgmtService\x12\x35\n\rserver_status\x12\r.shared.Empty\x1a\x15.mytorch.ServerStatus\x12\x33\n\x11\x63lient_disconnect\x12\r.shared.Empty\x1a\r.shared.Empty\"\x00\x12;\n\x14get_server_gpu_stats\x12\r.shared.Empty\x1a\x14.mytorch.GpuInfoListb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.scaffolding.server_mgmt_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SERVERSTATUS']._serialized_start=69
  _globals['_SERVERSTATUS']._serialized_end=167
  _globals['_GPUINFOLIST']._serialized_start=169
  _globals['_GPUINFOLIST']._serialized_end=227
  _globals['_SERVERMGMTSERVICE']._serialized_start=230
  _globals['_SERVERMGMTSERVICE']._serialized_end=418
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

# This is an abstract parent class for all datasets in the mytorchvision library.

from utils.logger import Logger

class Dataset:

    def __init__(self):
        self.uuid = None
        self.logger = Logger.get_logger()


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from .scaffolding import *
from .data_mgmt import folder_exists_on_server


__all__ = ['print_timing_data',
           'get_timing_data',
           'begin_run', 
           'end_run',
           'select_server', 
           'get_server_name', 
           'get_server', 
           'SERVER_LOCALHOST', 
           'SERVER_TITAN', 
           'SERVER_AWS',
           'SERVER_FUNC2',
           'SERVER_CACHEQ_AI',
           'folder_exists_on_server',
           'set_overwrite_server_data',
           'get_gpu_info'
           ]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytorch.nn import module_pb2 as mytorch_dot_nn_dot_module__pb2
from gRPC_impl.mytorch.nn import nn_msg_types_pb2 as mytorch_dot_nn_dot_nn__msg__types__pb2
from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytorch/nn/module_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ModuleServiceStub(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Forward = channel.unary_unary(
                '/mytorch.ModuleService/Forward',
                request_serializer=mytorch_dot_nn_dot_nn__msg__types__pb2.ForwardPassRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.CreateModuleOnServer = channel.unary_unary(
                '/mytorch.ModuleService/CreateModuleOnServer',
                request_serializer=shared__msg__types__pb2.Empty.SerializeToString,
                response_deserializer=shared__msg__types__pb2.UUID.FromString,
                _registered_method=True)
        self.GetStateDict = channel.unary_stream(
                '/mytorch.ModuleService/GetStateDict',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.NamedTensor.FromString,
                _registered_method=True)
        self.LoadStateDict = channel.stream_unary(
                '/mytorch.ModuleService/LoadStateDict',
                request_serializer=shared__msg__types__pb2.NamedTensor.SerializeToString,
                response_deserializer=shared__msg__types__pb2.BooleanValue.FromString,
                _registered_method=True)
        self.Eval = channel.unary_unary(
                '/mytorch.ModuleService/Eval',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.Empty.FromString,
                _registered_method=True)
        self.ToDevice = channel.unary_unary(
                '/mytorch.ModuleService/ToDevice',
                request_serializer=mytorch_dot_nn_dot_module__pb2.ModuleToDeviceRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.Empty.FromString,
                _registered_method=True)
        self.GetParameters = channel.unary_unary(
                '/mytorch.ModuleService/GetParameters',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=mytorch_dot_nn_dot_module__pb2.Parameters.FromString,
                _registered_method=True)
        self.half = channel.unary_unary(
                '/mytorch.ModuleService/half',
                request_serializer=mytorch_dot_nn_dot_module__pb2.half_request.SerializeToString,
                response_deserializer=mytorch_dot_nn_dot_module__pb2.half_response.FromString,
                _registered_method=True)
        self.delete = channel.unary_unary(
                '/mytorch.ModuleService/delete',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.Empty.FromString,
                _registered_method=True)


class ModuleServiceServicer(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def Forward(self, request, context):
        """model.forward() or model(x)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateModuleOnServer(self, request, context):
        """torch.nn.Module()
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStateDict(self, request, context):
        """rpc GetStateDict          (shared.UUID)                  returns (SerializedStateDict); // model.state_dict()
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LoadStateDict(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Eval(self, request, context):
        """model.eval()
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ToDevice(self, request, context):
        """model.to(device)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetParameters(self, request, context):
        """model.parameters()
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def half(self, request, context):
        """model.half()
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def delete(self, request, context):
        """removes object from ObjectStorage and deletes it
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModuleServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Forward': grpc.unary_unary_rpc_method_handler(
                    servicer.Forward,
                    request_deserializer=mytorch_dot_nn_dot_nn__msg__types__pb2.ForwardPassRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'CreateModuleOnServer': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateModuleOnServer,
                    request_deserializer=shared__msg__types__pb2.Empty.FromString,
                    response_serializer=shared__msg__types__pb2.UUID.SerializeToString,
            ),
            'GetStateDict': grpc.unary_stream_rpc_method_handler(
                    servicer.GetStateDict,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.NamedTensor.SerializeToString,
            ),
            'LoadStateDict': grpc.stream_unary_rpc_method_handler(
                    servicer.LoadStateDict,
                    request_deserializer=shared__msg__types__pb2.NamedTensor.FromString,
                    response_serializer=shared__msg__types__pb2.BooleanValue.SerializeToString,
            ),
            'Eval': grpc.unary_unary_rpc_method_handler(
                    servicer.Eval,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.Empty.SerializeToString,
            ),
            'ToDevice': grpc.unary_unary_rpc_method_handler(
                    servicer.ToDevice,
                    request_deserializer=mytorch_dot_nn_dot_module__pb2.ModuleToDeviceRequest.FromString,
                    response_serializer=shared__msg__types__pb2.Empty.SerializeToString,
            ),
            'GetParameters': grpc.unary_unary_rpc_method_handler(
                    servicer.GetParameters,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=mytorch_dot_nn_dot_module__pb2.Parameters.SerializeToString,
            ),
            'half': grpc.unary_unary_rpc_method_handler(
                    servicer.half,
                    request_deserializer=mytorch_dot_nn_dot_module__pb2.half_request.FromString,
                    response_serializer=mytorch_dot_nn_dot_module__pb2.half_response.SerializeToString,
            ),
            'delete': grpc.unary_unary_rpc_method_handler(
                    servicer.delete,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytorch.ModuleService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytorch.ModuleService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ModuleService(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    @staticmethod
    def Forward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/Forward',
            mytorch_dot_nn_dot_nn__msg__types__pb2.ForwardPassRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateModuleOnServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/CreateModuleOnServer',
            shared__msg__types__pb2.Empty.SerializeToString,
            shared__msg__types__pb2.UUID.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStateDict(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/mytorch.ModuleService/GetStateDict',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.NamedTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def LoadStateDict(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/mytorch.ModuleService/LoadStateDict',
            shared__msg__types__pb2.NamedTensor.SerializeToString,
            shared__msg__types__pb2.BooleanValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Eval(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/Eval',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ToDevice(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/ToDevice',
            mytorch_dot_nn_dot_module__pb2.ModuleToDeviceRequest.SerializeToString,
            shared__msg__types__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetParameters(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/GetParameters',
            shared__msg__types__pb2.UUID.SerializeToString,
            mytorch_dot_nn_dot_module__pb2.Parameters.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def half(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/half',
            mytorch_dot_nn_dot_module__pb2.half_request.SerializeToString,
            mytorch_dot_nn_dot_module__pb2.half_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def delete(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.ModuleService/delete',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ServerStatus(_message.Message):
    __slots__ = ("server_name", "server_ip", "server_port", "server_status")
    SERVER_NAME_FIELD_NUMBER: _ClassVar[int]
    SERVER_IP_FIELD_NUMBER: _ClassVar[int]
    SERVER_PORT_FIELD_NUMBER: _ClassVar[int]
    SERVER_STATUS_FIELD_NUMBER: _ClassVar[int]
    server_name: str
    server_ip: str
    server_port: int
    server_status: str
    def __init__(self, server_name: _Optional[str] = ..., server_ip: _Optional[str] = ..., server_port: _Optional[int] = ..., server_status: _Optional[str] = ...) -> None: ...

class GpuInfoList(_message.Message):
    __slots__ = ("gpu_dicts",)
    GPU_DICTS_FIELD_NUMBER: _ClassVar[int]
    gpu_dicts: _containers.RepeatedCompositeFieldContainer[_shared_msg_types_pb2.StringDictionary]
    def __init__(self, gpu_dicts: _Optional[_Iterable[_Union[_shared_msg_types_pb2.StringDictionary, _Mapping]]] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.jit import jit_pb2, jit_pb2_grpc
from gRPC_impl import shared_msg_types_pb2
from torch.nn.Module import Module
from torch.Tensor import Tensor
from utils.logger import Logger

class JitProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = jit_pb2_grpc.JitServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = None

    @wrap_with_error_handler
    def trace(self, model: Module, example_input: Tensor) -> str:
        # Load a model from the PyTorch Hub
        request = jit_pb2.ModelIDandTensorID(model_uuid=model.uuid, tensor_uuid=example_input.uuid)
        response = self.stub.trace(request)
        self.logger.info(f"Traced model has been generated")
        return response.uuid

    @wrap_with_error_handler
    def save(self, module: Module, path: str) -> None:
        request = jit_pb2.SaveRequest(model_uuid=module.uuid, path=path)
        self.stub.save(request)
        self.logger.info(f"Model has been saved to {path}")

    @wrap_with_error_handler
    def load(self, path: str) -> str:
        request = shared_msg_types_pb2.StringValue(value=path)
        response = self.stub.load(request)
        self.logger.info(f"Model has been loaded from {path}")
        return response.uuid


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.nn.layers_proxy import ReLUProxy

class ReLU:
    def __init__(self):
        self.proxy = ReLUProxy()

    def __call__(self, tensor):
        # Delegate the call to ReLUProxy for execution
        return self.proxy.forward(tensor)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.scaffolding.server_mgmt_proxy import ServerMgmtProxy
from connection_utils.server_connection import ServerConnection
from utils.logger import Logger

def server_status() -> str:
    if ServerConnection.is_connection_active():
        return ServerMgmtProxy().client_disconnect()
    else:
        return

def client_disconnect() -> None:
    Logger.get_logger().info("Disconnecting from server...")
    return ServerMgmtProxy().client_disconnect()

def print_server_gpu_stats() -> None:
    info_list = ServerMgmtProxy().get_server_gpu_stats()
    for index, info_dict in enumerate(info_list):
        print(f"~~~ GPU {index} ~~~")
        # print info_dict
        for key, value in info_dict.items():
            print(f"{key}: {value}")


from enum import Enum, unique

@unique
class WeightsEnum(Enum):
    DEFAULT = 0
    IMAGENET1K_V1 = 1
    IMAGENET1K_V2 = 2

# Dynamically create a new Enum that includes all members from WeightsEnum
# and potentially adds new ones
ResNet18_Weights = Enum('ResNet18_Weights', [(name, member.value) for name, member in WeightsEnum.__members__.items()])
ResNet50_Weights = Enum('ResNet50_Weights', [(name, member.value) for name, member in WeightsEnum.__members__.items()])
ResNet152_Weights = Enum('ResNet152_Weights', [(name, member.value) for name, member in WeightsEnum.__members__.items()])


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytorch import mytorch_pb2 as mytorch_dot_mytorch__pb2
from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytorch/mytorch_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class MyTorchServiceStub(object):
    """/////////////////////////////////////////////
    The MyTorch API / service definition
    /////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.randn = channel.unary_unary(
                '/mytorch.MyTorchService/randn',
                request_serializer=shared__msg__types__pb2.TensorShape.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.rand = channel.unary_unary(
                '/mytorch.MyTorchService/rand',
                request_serializer=shared__msg__types__pb2.TensorShape.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.from_numpy = channel.unary_unary(
                '/mytorch.MyTorchService/from_numpy',
                request_serializer=shared__msg__types__pb2.SerializedNumpyArray.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.arange = channel.unary_unary(
                '/mytorch.MyTorchService/arange',
                request_serializer=mytorch_dot_mytorch__pb2.ARangeRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.reshape = channel.unary_unary(
                '/mytorch.MyTorchService/reshape',
                request_serializer=shared__msg__types__pb2.ReshapeRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.cat = channel.unary_unary(
                '/mytorch.MyTorchService/cat',
                request_serializer=shared__msg__types__pb2.TensorIDsAndDim.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.argmax = channel.unary_unary(
                '/mytorch.MyTorchService/argmax',
                request_serializer=mytorch_dot_mytorch__pb2.ArgMaxRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.matmul = channel.unary_unary(
                '/mytorch.MyTorchService/matmul',
                request_serializer=shared__msg__types__pb2.TwoTensorIDs.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.sendPytorchTensor = channel.unary_unary(
                '/mytorch.MyTorchService/sendPytorchTensor',
                request_serializer=shared__msg__types__pb2.SerializedTensorObject.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.unsqueeze = channel.unary_unary(
                '/mytorch.MyTorchService/unsqueeze',
                request_serializer=shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.max = channel.unary_unary(
                '/mytorch.MyTorchService/max',
                request_serializer=shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
                response_deserializer=shared__msg__types__pb2.TwoGrpcTensors.FromString,
                _registered_method=True)
        self.meshgrid = channel.unary_unary(
                '/mytorch.MyTorchService/meshgrid',
                request_serializer=shared__msg__types__pb2.TensorIDsAndDim.SerializeToString,
                response_deserializer=shared__msg__types__pb2.MultipleGrpcTensors.FromString,
                _registered_method=True)
        self.repeat_interleave = channel.unary_unary(
                '/mytorch.MyTorchService/repeat_interleave',
                request_serializer=mytorch_dot_mytorch__pb2.RepeatInterleaveRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.allclose = channel.unary_unary(
                '/mytorch.MyTorchService/allclose',
                request_serializer=mytorch_dot_mytorch__pb2.AllCloseRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.BooleanValue.FromString,
                _registered_method=True)
        self.mytorch_server_get_timing_statistics = channel.unary_unary(
                '/mytorch.MyTorchService/mytorch_server_get_timing_statistics',
                request_serializer=shared__msg__types__pb2.StringValue.SerializeToString,
                response_deserializer=shared__msg__types__pb2.StringValue.FromString,
                _registered_method=True)
        self.mytorch_server_initialize_timing_statistics = channel.unary_unary(
                '/mytorch.MyTorchService/mytorch_server_initialize_timing_statistics',
                request_serializer=shared__msg__types__pb2.StringValue.SerializeToString,
                response_deserializer=shared__msg__types__pb2.BooleanValue.FromString,
                _registered_method=True)
        self.save_model_stream = channel.unary_stream(
                '/mytorch.MyTorchService/save_model_stream',
                request_serializer=mytorch_dot_mytorch__pb2.SaveModelRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.FileChunk.FromString,
                _registered_method=True)
        self.load_model_stream = channel.stream_unary(
                '/mytorch.MyTorchService/load_model_stream',
                request_serializer=shared__msg__types__pb2.FileChunk.SerializeToString,
                response_deserializer=mytorch_dot_mytorch__pb2.LoadModelResponse.FromString,
                _registered_method=True)
        self.generic_call = channel.unary_unary(
                '/mytorch.MyTorchService/generic_call',
                request_serializer=shared__msg__types__pb2.JsonRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.JsonResponse.FromString,
                _registered_method=True)


class MyTorchServiceServicer(object):
    """/////////////////////////////////////////////
    The MyTorch API / service definition
    /////////////////////////////////////////////
    """

    def randn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def rand(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def from_numpy(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def arange(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reshape(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def cat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def argmax(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def matmul(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendPytorchTensor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def unsqueeze(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def max(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def meshgrid(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def repeat_interleave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def allclose(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def mytorch_server_get_timing_statistics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def mytorch_server_initialize_timing_statistics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def save_model_stream(self, request, context):
        """DEPRECATED
        rpc save_model (SaveModelRequest) returns (SerializedModelFile);
        rpc load_model (SerializedModelFile) returns (LoadModelResponse);

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def load_model_stream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def generic_call(self, request, context):
        """Generic JSON-based method invocation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MyTorchServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'randn': grpc.unary_unary_rpc_method_handler(
                    servicer.randn,
                    request_deserializer=shared__msg__types__pb2.TensorShape.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'rand': grpc.unary_unary_rpc_method_handler(
                    servicer.rand,
                    request_deserializer=shared__msg__types__pb2.TensorShape.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'from_numpy': grpc.unary_unary_rpc_method_handler(
                    servicer.from_numpy,
                    request_deserializer=shared__msg__types__pb2.SerializedNumpyArray.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'arange': grpc.unary_unary_rpc_method_handler(
                    servicer.arange,
                    request_deserializer=mytorch_dot_mytorch__pb2.ARangeRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'reshape': grpc.unary_unary_rpc_method_handler(
                    servicer.reshape,
                    request_deserializer=shared__msg__types__pb2.ReshapeRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'cat': grpc.unary_unary_rpc_method_handler(
                    servicer.cat,
                    request_deserializer=shared__msg__types__pb2.TensorIDsAndDim.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'argmax': grpc.unary_unary_rpc_method_handler(
                    servicer.argmax,
                    request_deserializer=mytorch_dot_mytorch__pb2.ArgMaxRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'matmul': grpc.unary_unary_rpc_method_handler(
                    servicer.matmul,
                    request_deserializer=shared__msg__types__pb2.TwoTensorIDs.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'sendPytorchTensor': grpc.unary_unary_rpc_method_handler(
                    servicer.sendPytorchTensor,
                    request_deserializer=shared__msg__types__pb2.SerializedTensorObject.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'unsqueeze': grpc.unary_unary_rpc_method_handler(
                    servicer.unsqueeze,
                    request_deserializer=shared__msg__types__pb2.TensorIDAndDim.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'max': grpc.unary_unary_rpc_method_handler(
                    servicer.max,
                    request_deserializer=shared__msg__types__pb2.TensorIDAndDim.FromString,
                    response_serializer=shared__msg__types__pb2.TwoGrpcTensors.SerializeToString,
            ),
            'meshgrid': grpc.unary_unary_rpc_method_handler(
                    servicer.meshgrid,
                    request_deserializer=shared__msg__types__pb2.TensorIDsAndDim.FromString,
                    response_serializer=shared__msg__types__pb2.MultipleGrpcTensors.SerializeToString,
            ),
            'repeat_interleave': grpc.unary_unary_rpc_method_handler(
                    servicer.repeat_interleave,
                    request_deserializer=mytorch_dot_mytorch__pb2.RepeatInterleaveRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'allclose': grpc.unary_unary_rpc_method_handler(
                    servicer.allclose,
                    request_deserializer=mytorch_dot_mytorch__pb2.AllCloseRequest.FromString,
                    response_serializer=shared__msg__types__pb2.BooleanValue.SerializeToString,
            ),
            'mytorch_server_get_timing_statistics': grpc.unary_unary_rpc_method_handler(
                    servicer.mytorch_server_get_timing_statistics,
                    request_deserializer=shared__msg__types__pb2.StringValue.FromString,
                    response_serializer=shared__msg__types__pb2.StringValue.SerializeToString,
            ),
            'mytorch_server_initialize_timing_statistics': grpc.unary_unary_rpc_method_handler(
                    servicer.mytorch_server_initialize_timing_statistics,
                    request_deserializer=shared__msg__types__pb2.StringValue.FromString,
                    response_serializer=shared__msg__types__pb2.BooleanValue.SerializeToString,
            ),
            'save_model_stream': grpc.unary_stream_rpc_method_handler(
                    servicer.save_model_stream,
                    request_deserializer=mytorch_dot_mytorch__pb2.SaveModelRequest.FromString,
                    response_serializer=shared__msg__types__pb2.FileChunk.SerializeToString,
            ),
            'load_model_stream': grpc.stream_unary_rpc_method_handler(
                    servicer.load_model_stream,
                    request_deserializer=shared__msg__types__pb2.FileChunk.FromString,
                    response_serializer=mytorch_dot_mytorch__pb2.LoadModelResponse.SerializeToString,
            ),
            'generic_call': grpc.unary_unary_rpc_method_handler(
                    servicer.generic_call,
                    request_deserializer=shared__msg__types__pb2.JsonRequest.FromString,
                    response_serializer=shared__msg__types__pb2.JsonResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytorch.MyTorchService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytorch.MyTorchService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class MyTorchService(object):
    """/////////////////////////////////////////////
    The MyTorch API / service definition
    /////////////////////////////////////////////
    """

    @staticmethod
    def randn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/randn',
            shared__msg__types__pb2.TensorShape.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def rand(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/rand',
            shared__msg__types__pb2.TensorShape.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def from_numpy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/from_numpy',
            shared__msg__types__pb2.SerializedNumpyArray.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def arange(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/arange',
            mytorch_dot_mytorch__pb2.ARangeRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def reshape(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/reshape',
            shared__msg__types__pb2.ReshapeRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def cat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/cat',
            shared__msg__types__pb2.TensorIDsAndDim.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def argmax(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/argmax',
            mytorch_dot_mytorch__pb2.ArgMaxRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def matmul(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/matmul',
            shared__msg__types__pb2.TwoTensorIDs.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def sendPytorchTensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/sendPytorchTensor',
            shared__msg__types__pb2.SerializedTensorObject.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def unsqueeze(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/unsqueeze',
            shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def max(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/max',
            shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
            shared__msg__types__pb2.TwoGrpcTensors.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def meshgrid(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/meshgrid',
            shared__msg__types__pb2.TensorIDsAndDim.SerializeToString,
            shared__msg__types__pb2.MultipleGrpcTensors.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def repeat_interleave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/repeat_interleave',
            mytorch_dot_mytorch__pb2.RepeatInterleaveRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def allclose(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/allclose',
            mytorch_dot_mytorch__pb2.AllCloseRequest.SerializeToString,
            shared__msg__types__pb2.BooleanValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def mytorch_server_get_timing_statistics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/mytorch_server_get_timing_statistics',
            shared__msg__types__pb2.StringValue.SerializeToString,
            shared__msg__types__pb2.StringValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def mytorch_server_initialize_timing_statistics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/mytorch_server_initialize_timing_statistics',
            shared__msg__types__pb2.StringValue.SerializeToString,
            shared__msg__types__pb2.BooleanValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def save_model_stream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/mytorch.MyTorchService/save_model_stream',
            mytorch_dot_mytorch__pb2.SaveModelRequest.SerializeToString,
            shared__msg__types__pb2.FileChunk.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def load_model_stream(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/mytorch.MyTorchService/load_model_stream',
            shared__msg__types__pb2.FileChunk.SerializeToString,
            mytorch_dot_mytorch__pb2.LoadModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def generic_call(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.MyTorchService/generic_call',
            shared__msg__types__pb2.JsonRequest.SerializeToString,
            shared__msg__types__pb2.JsonResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SerializedTensorObject(_message.Message):
    __slots__ = ("serialized_tensor",)
    SERIALIZED_TENSOR_FIELD_NUMBER: _ClassVar[int]
    serialized_tensor: bytes
    def __init__(self, serialized_tensor: _Optional[bytes] = ...) -> None: ...

class SerializedNumpyArray(_message.Message):
    __slots__ = ("data", "shape", "dtype")
    DATA_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    def __init__(self, data: _Optional[bytes] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ...) -> None: ...

class SerializedTensorData(_message.Message):
    __slots__ = ("data", "shape", "dtype")
    DATA_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    def __init__(self, data: _Optional[bytes] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ...) -> None: ...

class IntegerValue(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    def __init__(self, value: _Optional[int] = ...) -> None: ...

class FloatValue(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: float
    def __init__(self, value: _Optional[float] = ...) -> None: ...

class StringValue(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    def __init__(self, value: _Optional[str] = ...) -> None: ...

class StringPair(_message.Message):
    __slots__ = ("key", "value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class ReshapeRequest(_message.Message):
    __slots__ = ("tensor_uuid", "shape")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    shape: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, tensor_uuid: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...

class Empty(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class BooleanValue(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: bool
    def __init__(self, value: bool = ...) -> None: ...

class StringDictionary(_message.Message):
    __slots__ = ("items",)
    class ItemsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: _containers.ScalarMap[str, str]
    def __init__(self, items: _Optional[_Mapping[str, str]] = ...) -> None: ...

class keyValuePair(_message.Message):
    __slots__ = ("key", "string_value", "int_value", "float_value", "bool_value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    string_value: str
    int_value: int
    float_value: float
    bool_value: bool
    def __init__(self, key: _Optional[str] = ..., string_value: _Optional[str] = ..., int_value: _Optional[int] = ..., float_value: _Optional[float] = ..., bool_value: bool = ...) -> None: ...

class UUID(_message.Message):
    __slots__ = ("uuid",)
    UUID_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    def __init__(self, uuid: _Optional[str] = ...) -> None: ...

class Dtype(_message.Message):
    __slots__ = ("dtype",)
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    dtype: str
    def __init__(self, dtype: _Optional[str] = ...) -> None: ...

class TensorIDAndDim(_message.Message):
    __slots__ = ("tensor_uuid", "dim")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    DIM_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    dim: int
    def __init__(self, tensor_uuid: _Optional[str] = ..., dim: _Optional[int] = ...) -> None: ...

class TensorIDsAndDim(_message.Message):
    __slots__ = ("tensor_uuids", "dim")
    TENSOR_UUIDS_FIELD_NUMBER: _ClassVar[int]
    DIM_FIELD_NUMBER: _ClassVar[int]
    tensor_uuids: _containers.RepeatedScalarFieldContainer[str]
    dim: int
    def __init__(self, tensor_uuids: _Optional[_Iterable[str]] = ..., dim: _Optional[int] = ...) -> None: ...

class TwoTensorIDs(_message.Message):
    __slots__ = ("tensor1_uuid", "tensor2_uuid")
    TENSOR1_UUID_FIELD_NUMBER: _ClassVar[int]
    TENSOR2_UUID_FIELD_NUMBER: _ClassVar[int]
    tensor1_uuid: str
    tensor2_uuid: str
    def __init__(self, tensor1_uuid: _Optional[str] = ..., tensor2_uuid: _Optional[str] = ...) -> None: ...

class TensorShape(_message.Message):
    __slots__ = ("shape",)
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    shape: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, shape: _Optional[_Iterable[int]] = ...) -> None: ...

class GrpcTensor(_message.Message):
    __slots__ = ("uuid", "shape", "dtype")
    UUID_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    def __init__(self, uuid: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ...) -> None: ...

class NamedTensor(_message.Message):
    __slots__ = ("name", "tensor")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TENSOR_FIELD_NUMBER: _ClassVar[int]
    name: str
    tensor: GrpcTensor
    def __init__(self, name: _Optional[str] = ..., tensor: _Optional[_Union[GrpcTensor, _Mapping]] = ...) -> None: ...

class TwoGrpcTensors(_message.Message):
    __slots__ = ("tensor1", "tensor2")
    TENSOR1_FIELD_NUMBER: _ClassVar[int]
    TENSOR2_FIELD_NUMBER: _ClassVar[int]
    tensor1: GrpcTensor
    tensor2: GrpcTensor
    def __init__(self, tensor1: _Optional[_Union[GrpcTensor, _Mapping]] = ..., tensor2: _Optional[_Union[GrpcTensor, _Mapping]] = ...) -> None: ...

class MultipleGrpcTensors(_message.Message):
    __slots__ = ("tensors",)
    TENSORS_FIELD_NUMBER: _ClassVar[int]
    tensors: _containers.RepeatedCompositeFieldContainer[GrpcTensor]
    def __init__(self, tensors: _Optional[_Iterable[_Union[GrpcTensor, _Mapping]]] = ...) -> None: ...

class Value(_message.Message):
    __slots__ = ("string_value", "int_value", "float_value", "bool_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    int_value: int
    float_value: float
    bool_value: bool
    def __init__(self, string_value: _Optional[str] = ..., int_value: _Optional[int] = ..., float_value: _Optional[float] = ..., bool_value: bool = ...) -> None: ...

class FileChunk(_message.Message):
    __slots__ = ("filename", "chunk_data", "chunk_index", "is_last_chunk")
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    CHUNK_DATA_FIELD_NUMBER: _ClassVar[int]
    CHUNK_INDEX_FIELD_NUMBER: _ClassVar[int]
    IS_LAST_CHUNK_FIELD_NUMBER: _ClassVar[int]
    filename: str
    chunk_data: bytes
    chunk_index: int
    is_last_chunk: bool
    def __init__(self, filename: _Optional[str] = ..., chunk_data: _Optional[bytes] = ..., chunk_index: _Optional[int] = ..., is_last_chunk: bool = ...) -> None: ...

class JsonRequest(_message.Message):
    __slots__ = ("json_payload",)
    JSON_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    json_payload: str
    def __init__(self, json_payload: _Optional[str] = ...) -> None: ...

class JsonResponse(_message.Message):
    __slots__ = ("json_payload",)
    JSON_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    json_payload: str
    def __init__(self, json_payload: _Optional[str] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytorch import tensor_pb2 as mytorch_dot_tensor__pb2
from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytorch/tensor_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class TensorServiceStub(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.float = channel.unary_unary(
                '/mytorch.TensorService/float',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.long = channel.unary_unary(
                '/mytorch.TensorService/long',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.backward = channel.unary_unary(
                '/mytorch.TensorService/backward',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.to_cuda = channel.unary_unary(
                '/mytorch.TensorService/to_cuda',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.to_cpu = channel.unary_unary(
                '/mytorch.TensorService/to_cpu',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.sum = channel.unary_unary(
                '/mytorch.TensorService/sum',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.add = channel.unary_unary(
                '/mytorch.TensorService/add',
                request_serializer=shared__msg__types__pb2.TwoTensorIDs.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.equal = channel.unary_unary(
                '/mytorch.TensorService/equal',
                request_serializer=shared__msg__types__pb2.TwoTensorIDs.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.index = channel.unary_unary(
                '/mytorch.TensorService/index',
                request_serializer=shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.unsqueeze = channel.unary_unary(
                '/mytorch.TensorService/unsqueeze',
                request_serializer=shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.reshape = channel.unary_unary(
                '/mytorch.TensorService/reshape',
                request_serializer=shared__msg__types__pb2.ReshapeRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.slice = channel.unary_unary(
                '/mytorch.TensorService/slice',
                request_serializer=mytorch_dot_tensor__pb2.SliceRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.GrpcTensor.FromString,
                _registered_method=True)
        self.get_data = channel.unary_unary(
                '/mytorch.TensorService/get_data',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.SerializedTensorData.FromString,
                _registered_method=True)
        self.item = channel.unary_unary(
                '/mytorch.TensorService/item',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.FloatValue.FromString,
                _registered_method=True)
        self.get_shape = channel.unary_unary(
                '/mytorch.TensorService/get_shape',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.TensorShape.FromString,
                _registered_method=True)
        self.get_dtype = channel.unary_unary(
                '/mytorch.TensorService/get_dtype',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.Dtype.FromString,
                _registered_method=True)
        self.delete = channel.unary_unary(
                '/mytorch.TensorService/delete',
                request_serializer=shared__msg__types__pb2.UUID.SerializeToString,
                response_deserializer=shared__msg__types__pb2.Empty.FromString,
                _registered_method=True)
        self.TensorStreamingOp = channel.stream_stream(
                '/mytorch.TensorService/TensorStreamingOp',
                request_serializer=mytorch_dot_tensor__pb2.TensorStreamingRequest.SerializeToString,
                response_deserializer=mytorch_dot_tensor__pb2.TensorStreamingResponse.FromString,
                _registered_method=True)
        self.generic_call = channel.unary_unary(
                '/mytorch.TensorService/generic_call',
                request_serializer=shared__msg__types__pb2.JsonRequest.SerializeToString,
                response_deserializer=shared__msg__types__pb2.JsonResponse.FromString,
                _registered_method=True)


class TensorServiceServicer(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def float(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def long(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def backward(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def to_cuda(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def to_cpu(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sum(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def add(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def equal(self, request, context):
        """returns a boolean tensor
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def index(self, request, context):
        """handles indexing, e.g. tensor[0]
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def unsqueeze(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reshape(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def slice(self, request, context):
        """handles indexing and slicing, e.g. tensor[0:2, 0:2, 5:]
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_data(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def item(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_shape(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_dtype(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def delete(self, request, context):
        """removes object from ObjectStorage and deletes it
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TensorStreamingOp(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def generic_call(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TensorServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'float': grpc.unary_unary_rpc_method_handler(
                    servicer.float,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'long': grpc.unary_unary_rpc_method_handler(
                    servicer.long,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'backward': grpc.unary_unary_rpc_method_handler(
                    servicer.backward,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'to_cuda': grpc.unary_unary_rpc_method_handler(
                    servicer.to_cuda,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'to_cpu': grpc.unary_unary_rpc_method_handler(
                    servicer.to_cpu,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'sum': grpc.unary_unary_rpc_method_handler(
                    servicer.sum,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'add': grpc.unary_unary_rpc_method_handler(
                    servicer.add,
                    request_deserializer=shared__msg__types__pb2.TwoTensorIDs.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'equal': grpc.unary_unary_rpc_method_handler(
                    servicer.equal,
                    request_deserializer=shared__msg__types__pb2.TwoTensorIDs.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'index': grpc.unary_unary_rpc_method_handler(
                    servicer.index,
                    request_deserializer=shared__msg__types__pb2.TensorIDAndDim.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'unsqueeze': grpc.unary_unary_rpc_method_handler(
                    servicer.unsqueeze,
                    request_deserializer=shared__msg__types__pb2.TensorIDAndDim.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'reshape': grpc.unary_unary_rpc_method_handler(
                    servicer.reshape,
                    request_deserializer=shared__msg__types__pb2.ReshapeRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'slice': grpc.unary_unary_rpc_method_handler(
                    servicer.slice,
                    request_deserializer=mytorch_dot_tensor__pb2.SliceRequest.FromString,
                    response_serializer=shared__msg__types__pb2.GrpcTensor.SerializeToString,
            ),
            'get_data': grpc.unary_unary_rpc_method_handler(
                    servicer.get_data,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.SerializedTensorData.SerializeToString,
            ),
            'item': grpc.unary_unary_rpc_method_handler(
                    servicer.item,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.FloatValue.SerializeToString,
            ),
            'get_shape': grpc.unary_unary_rpc_method_handler(
                    servicer.get_shape,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.TensorShape.SerializeToString,
            ),
            'get_dtype': grpc.unary_unary_rpc_method_handler(
                    servicer.get_dtype,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.Dtype.SerializeToString,
            ),
            'delete': grpc.unary_unary_rpc_method_handler(
                    servicer.delete,
                    request_deserializer=shared__msg__types__pb2.UUID.FromString,
                    response_serializer=shared__msg__types__pb2.Empty.SerializeToString,
            ),
            'TensorStreamingOp': grpc.stream_stream_rpc_method_handler(
                    servicer.TensorStreamingOp,
                    request_deserializer=mytorch_dot_tensor__pb2.TensorStreamingRequest.FromString,
                    response_serializer=mytorch_dot_tensor__pb2.TensorStreamingResponse.SerializeToString,
            ),
            'generic_call': grpc.unary_unary_rpc_method_handler(
                    servicer.generic_call,
                    request_deserializer=shared__msg__types__pb2.JsonRequest.FromString,
                    response_serializer=shared__msg__types__pb2.JsonResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytorch.TensorService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytorch.TensorService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class TensorService(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    @staticmethod
    def float(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/float',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def long(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/long',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def backward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/backward',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def to_cuda(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/to_cuda',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def to_cpu(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/to_cpu',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def sum(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/sum',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def add(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/add',
            shared__msg__types__pb2.TwoTensorIDs.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def equal(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/equal',
            shared__msg__types__pb2.TwoTensorIDs.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def index(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/index',
            shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def unsqueeze(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/unsqueeze',
            shared__msg__types__pb2.TensorIDAndDim.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def reshape(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/reshape',
            shared__msg__types__pb2.ReshapeRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def slice(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/slice',
            mytorch_dot_tensor__pb2.SliceRequest.SerializeToString,
            shared__msg__types__pb2.GrpcTensor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def get_data(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/get_data',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.SerializedTensorData.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def item(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/item',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.FloatValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def get_shape(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/get_shape',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.TensorShape.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def get_dtype(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/get_dtype',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.Dtype.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def delete(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/delete',
            shared__msg__types__pb2.UUID.SerializeToString,
            shared__msg__types__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TensorStreamingOp(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/mytorch.TensorService/TensorStreamingOp',
            mytorch_dot_tensor__pb2.TensorStreamingRequest.SerializeToString,
            mytorch_dot_tensor__pb2.TensorStreamingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def generic_call(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.TensorService/generic_call',
            shared__msg__types__pb2.JsonRequest.SerializeToString,
            shared__msg__types__pb2.JsonResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
